import { ISerializeObject, ISerializer, Serialize } from './interface'
import { isObject, isPlainObject, PrimitivePrototype } from '../utils'
import { GlobalRegistry, Registry } from '../registry'

class UnhandledTypeError extends TypeError {
  constructor(serializer: ISerializer<unknown>, value: any) {
    super(`unhandled type: '${typeof value}' for serializer ${serializer.constructor.name}`)
  }
}

function serializeObject<T, U extends object, S extends ISerializeObject<T>>(serializer: S, obj: U): T {
  for (const key in obj) {
    const value = obj[key]
    serializer.serializeEntry(key, value)
  }

  return serializer.end()
}

function serializeClass<T, U extends object, S extends ISerializer<T>>(serializer: S, value: U): T {
  const name = value.constructor.name
  const ser = serializer.serializeClass(name)
  return serializeObject(ser, value)
}

function getSerialize<T, U>(value: U, registry: Registry): Serialize<T> {
  if (isObject(value)) {
    return registry.serializers.get(value.constructor) as Serialize<T>
  }

  return registry.serializers.get(PrimitivePrototype[typeof value]) || defaultSerialize as Serialize<T>
}

function defaultSerialize<T, U, S extends ISerializer<T>>(serializer: S, value: U): T {
  switch (typeof value) {
    case 'string': return serializer.serializeString(value)
    case 'number': return serializer.serializeNumber(value)
    case 'bigint': return serializer.serializeBigInt(value)
    case 'boolean': return serializer.serializeBoolean(value)
    case 'symbol': return serializer.serializeSymbol(value)
    case 'undefined': return serializer.serializeNull()
    case 'object':
      switch (true) {
        case value == null: return serializer.serializeNull()
        case !isPlainObject(value): return serializeClass(serializer, value)
        default: return serializeObject(serializer.serializeObject(), value)
      }
    default: throw new UnhandledTypeError(serializer, value)
  }
}

export function serialize<T, U, S extends ISerializer<T>>(serializer: S, value: U, registry: Registry = GlobalRegistry): T {
  const ser = getSerialize<T, U>(value, registry)
  return ser(serializer, value)
}

